#include "LmerGraph.h"
#include "split.h"
#include <iostream>
#include <cmath>
#include <fstream>
#include <string>
#include <sstream>
#include <vector>


LmerGraph::LmerGraph(string _fin_counts, int _n_of_lmers, 
    int _max_n_of_neighbors, string _f_reads,
    int _count_threshold, int _min_unassigned_lmers, 
    int _min_number_of_neighbors, 
    int _max_number_of_nodes_to_add, string _output_lmer_clusters, 
    int _min_counts, int _n_of_reads,
    int _read_length)
{
    fin_counts = _fin_counts;
    n_of_nodes = 0;
    n_of_lmers = _n_of_lmers;
    max_n_of_neighbors = _max_n_of_neighbors;
    lmer_counts = new int [n_of_lmers];
    lmer_to_node = new int [n_of_lmers];
    for (int i = 0; i < n_of_lmers; i++)
    {
        lmer_counts[i] = 0;
        lmer_to_node[i] = -1;
    }
    f_reads = _f_reads;
    count_threshold = _count_threshold;
    min_unassigned_lmers = _min_unassigned_lmers;
    min_number_of_neighbors = _min_number_of_neighbors; 
    max_number_of_nodes_to_add = _max_number_of_nodes_to_add;
    output_lmer_clusters = _output_lmer_clusters;
    min_counts = _min_counts;
    n_of_reads = _n_of_reads;
    read_length =  _read_length;
}
void LmerGraph::fillLmers()
{    
    ifstream file_lmers_freq; 
    file_lmers_freq.open(fin_counts.c_str());
    string cur_s;
    int counter = 0;
    while(getline(file_lmers_freq,cur_s))
    {
        vector<string> line = split(cur_s, ' ');
        int index = atoi(line[0].c_str());
        int value = atoi(line[1].c_str()); 
        lmer_counts[index - 1] = value;
        if (value > 1)
        {
            lmer_to_node[index - 1] = counter;
            counter ++;
        }
    }
    n_of_nodes = counter;   
    file_lmers_freq.close();       
}

void LmerGraph::fillNodes()
{    
    node_to_lmer = new int [n_of_nodes];
    node_neighbors = new int [n_of_nodes];
    node_is_out = new bool [n_of_nodes];
    node_cluster = new int [n_of_nodes];
    for (int i = 0; i < n_of_lmers; i++)
    {
        int j = lmer_to_node[i];
        if (j >= 0)  
             node_to_lmer[j] = i;      
    } 
    for (int i = 0; i < n_of_nodes; i++)
    {
        node_neighbors[i] = 0; 
        node_is_out[i] = false;
        node_cluster[i] = 0;
    }     
}

void LmerGraph::initGraph()
{
    neighbors = new int * [n_of_nodes];
    for(int i = 0; i < n_of_nodes; i++) 
        neighbors[i] = new int [max_n_of_neighbors];
    for(int i = 0; i < n_of_nodes; i++)
        for(int j = 0; j < max_n_of_neighbors; j++ )  
            neighbors[i][j] = -1;  
}

void LmerGraph::printProgress(int &reads_counter,double &persent)
{
    reads_counter++;
    double cur_percent = (100.0 * reads_counter) / n_of_reads;
    if (cur_percent >= persent)
    {
        cout << persent << "%... " << endl;
        persent += 20;
    }
} 

void LmerGraph::fillGraph()
{
    ifstream file_reads; 
    cout << "Phase I. Building graph of unique l-mers." <<
        " It will take a long time... Persent completed: " << endl;
    set<int> clique;
    file_reads.open(f_reads.c_str());  
    string cur_s;
    int reads_counter = 0;
    double persent = 20;
    while(getline(file_reads,cur_s) )
    {
        if (cur_s[0] != '>')
        {
            vector<string> line = split(cur_s, ' ');
            int lmer = atoi(line[0].c_str());
            if (lmer_counts[lmer] > min_counts && lmer_to_node[lmer] > -1)	
                clique.insert(lmer_to_node[lmer]);                     	
        }		
        else
        { 							
            addClique(clique);
            clique.clear();
            printProgress(reads_counter,persent); 
        }
    }
    cout << "Phase I. Updating the graph..." << endl;
    refineGraph();
    file_reads.close();
}

void LmerGraph::addClique(set<int> &clique)
{
    set <int>::iterator it, it2; 
    int lmer1 ,lmer2;
    for (it = clique.begin(); it != clique.end(); it++)
    {
        lmer1 = *it;
        for (it2 = it; it2 != clique.end(); it2++)
        {
            lmer2 = *it2;
            if (lmer1 != lmer2)
            {
                addPair(lmer1,lmer2);
                addPair(lmer2,lmer1);
            }	
        }
    }
}

bool LmerGraph::isNeighbor(int lmer1,int lmer2)
{
    for(int i = 0; i < node_neighbors[lmer1]; i++)
        if (neighbors[lmer1][i] == lmer2)
            return true;
    return false;
}

bool LmerGraph::isFull(int lmer1)
{
    if (node_neighbors[lmer1] == max_n_of_neighbors)
        return true; 
    return false;
}

void LmerGraph::addNeighbor(int lmer1,int lmer2)
{
    neighbors[lmer1][node_neighbors[lmer1]] = lmer2;
    node_neighbors[lmer1] ++;
}

void LmerGraph::addPair(int lmer1, int lmer2)
{
    if (!isNeighbor(lmer1,lmer2))
        if(!isFull(lmer1))
            addNeighbor(lmer1,lmer2);
        else
            node_is_out[lmer1] = true;
}

void LmerGraph::refineGraph()
{
    for (int i = 0; i < n_of_nodes; i++)
        if (lmer_counts[i] > count_threshold && lmer_counts[i] <= min_counts)
            node_is_out[i] = true;   
}


bool LmerGraph::findUnassignedRead(set <int> &initialized_cluster, 
    ifstream &file_reads)
{ 
    initialized_cluster.clear();
    string cur_s;
    int counter_of_unassigned_lmers = 0;
    
    while(getline(file_reads,cur_s) )
    {
        if (cur_s[0] == '>')
        {
            processed_reads++;
            if (counter_of_unassigned_lmers > min_unassigned_lmers)               
                return true;
            initialized_cluster.clear();
            counter_of_unassigned_lmers = 0;
        }
        else
        {
            vector<string> line = split(cur_s, ' ');
            int index = atoi(line[0].c_str());
            int node = lmer_to_node[index];
            if (candidateNode(node))
            {
                initialized_cluster.insert(node);
                counter_of_unassigned_lmers ++; 
            }
        }
    }
    return false;
}

bool LmerGraph::candidateNode(int node)
{
    if (node == -1) 
        return false;
    if (node_is_out[node] == false && node_cluster[node] == 0)
        return true;
    return false;
}

void LmerGraph::candidateNodes(int node,multiset <int> &candidates_to_add,
    set<int> &nodes_just_added,set<int> &nodes_in_cluster)
{
    for(int i = 0; i < node_neighbors[node]; i++)
    {
        int new_node = neighbors[node][i];
        if (candidateNode(new_node) && nodes_just_added.count(new_node) == 0 
          && nodes_in_cluster.count(new_node) == 0)
            candidates_to_add.insert(new_node);    
    }   
}
void LmerGraph::newNodes(multiset <int> &candidates_to_add,
    set <int> &nodes_to_add)
{
    multiset<int>:: iterator it;  
    for (it = candidates_to_add.begin(); it != candidates_to_add.end(); it++)
        if (candidates_to_add.count(*it) >= min_number_of_neighbors)
            nodes_to_add.insert(*it);     
}

void LmerGraph::removeFromCandidates(int node)
{
    node_is_out[node] = true;
}

void LmerGraph::stopCluster(set <int> &nodes_to_add, set<int> &nodes_just_added)
{
    set<int>:: iterator it;
    for (it = nodes_just_added.begin(); it != nodes_just_added.end(); it++)
        removeFromCandidates(*it); 
    nodes_just_added.clear(); 
    for (it = nodes_to_add.begin(); it != nodes_to_add.end(); it++)
        removeFromCandidates(*it); 
    nodes_to_add.clear(); 
}

int LmerGraph::updateCluster(set<int> &nodes_to_add, 
    set<int> &nodes_just_added, set<int> &nodes_in_cluster)
{
    set<int>:: iterator it;
    for (it = nodes_just_added.begin(); it != nodes_just_added.end(); it++)
        nodes_in_cluster.insert(*it);
    nodes_just_added.clear(); 
    int added_nodes = nodes_to_add.size();
    for (it = nodes_to_add.begin(); it != nodes_to_add.end(); it++)
        nodes_just_added.insert(*it); 
    nodes_to_add.clear();
    return added_nodes; 
}


int LmerGraph::iteration(set<int> &nodes_in_cluster, set<int> &nodes_just_added)
{
    set<int>:: iterator it; 
    multiset <int> candidates_to_add;
    set <int> nodes_to_add;
    for (it = nodes_just_added.begin(); it != nodes_just_added.end(); it++)
    {
        candidateNodes(*it,candidates_to_add,nodes_just_added,nodes_in_cluster);   
    }
    newNodes(candidates_to_add,nodes_to_add);
    if (nodes_to_add.size() > max_number_of_nodes_to_add)
    {
        //cout << "Min_number_of_neighbors = " << 
        //      min_number_of_neighbors << " is too little" << endl; 
        if (validNextParameters())
        {
            //cout << "Taking min_number_of_neighbors = " << 
            //    min_number_of_neighbors + 1 << endl; 
            nodes_to_add.clear();
            nextParameters(); //cin >> a;
            return 1;
        }
        else
        {
            //cout << "Maximum of min_number_of_neighbors is reached." << 
            //  " Stopping cluster." << endl; 
            stopCluster(nodes_to_add, nodes_just_added);
            resetParameters(); 
            return 0; 
        }
    }    
    int added_nodes = updateCluster(nodes_to_add, nodes_just_added, 
       nodes_in_cluster);
    //cout << "Parameter (min_number_of_neighb) = " << min_number_of_neighbors <<
       // " is perfect. Added "  << added_nodes << "." << endl; 
    resetParameters(); 
    return added_nodes;
}

void LmerGraph::addNewCluster(set <int> &nodes_in_cluster, int current_cluster)
{
    set<int>:: iterator it;
    for (it = nodes_in_cluster.begin(); it != nodes_in_cluster.end(); it++)
        node_cluster[*it] = current_cluster;
} 


bool LmerGraph::validNextParameters()
{
    if (read_length - min_number_of_neighbors - 1 <= 0 ) return false;
    return true;
}

void LmerGraph::nextParameters()
{
    min_number_of_neighbors ++; 
    max_number_of_nodes_to_add = 2 * (read_length - min_number_of_neighbors);
}

void LmerGraph::resetParameters()
{
    min_number_of_neighbors = 1; 
    max_number_of_nodes_to_add = 2 * (read_length - min_number_of_neighbors);
}

void LmerGraph::growingClsuter(set <int> &initialized_cluster)
{
    set<int> nodes_in_cluster;
    //cout << "Read for initialization is " << processed_reads << endl; 
    int added_nodes = initialized_cluster.size();
    //cout << "Initially there are as many nodes: " << added_nodes << endl; 
    resetParameters();
    while (added_nodes)
    {        
        added_nodes = iteration(nodes_in_cluster,initialized_cluster);
    }  
    number_of_clusters ++;
    //cout << "cluster has " << nodes_in_cluster.size() << " nodes"  << endl;
    addNewCluster(nodes_in_cluster,number_of_clusters); 
}

void LmerGraph::outputClusters()
{ 
    ofstream file_out; 
    file_out.open(output_lmer_clusters.c_str());
    file_out << number_of_clusters << " clusters" << endl;
    for (int i = 0; i < n_of_nodes; i ++)
    {
        int lmer = node_to_lmer[i];
        int cluster = node_cluster[i];
        file_out << lmer + 1 << " " << cluster << endl; 
    }
    file_out.close();
} 

void LmerGraph::show_graph()
{
    int counter = 0;
    for (int j = 0; j < n_of_nodes; j++)
        if (node_is_out[j])
            counter++; 
    cout << "As many nodes are out after graph construction " << counter << 
        ". And total number of nodes is " << n_of_nodes << "." << endl;
    int i = 0;
    while (i!= -1) 
    {
        cout << "Printing row " << i <<  " (" << node_to_lmer[i] << ") " << endl;
        for (int j = 0; j < node_neighbors[i]; j++)
            cout << neighbors[i][j] << " (" << node_to_lmer[neighbors[i][j]] <<
             ") ";    
        cout << endl;
        cout << "Enter the row you want to see: ";
        cin >> i;
    }
}

void LmerGraph::outputLmerToNode()
{ 
    ofstream file_out; 
    string fout = output_lmer_clusters + ".LtoN";
    file_out.open(fout.c_str());
    for (int i = 0; i < n_of_lmers; i ++)
    {
        file_out << i+1 << " " << lmer_to_node[i] << endl; 
    }
    file_out.close();
} 

void LmerGraph::outputNodeToLmer()
{ 
    ofstream file_out; 
    string fout = output_lmer_clusters + ".NtoL";
    file_out.open(fout.c_str());
    for (int i = 0; i < n_of_nodes; i ++)
    {
        file_out << i << " " << node_to_lmer[i] + 1 << endl; 
    }
    file_out.close();
} 


void LmerGraph::phaseI()
{
    fillLmers();
    fillNodes();
   // outputLmerToNode();
   // outputNodeToLmer();
    initGraph();
    int counter = 0;
    for (int j = 0; j < n_of_nodes; j++)
        if (node_is_out[j])
            counter++; 
    fillGraph();

   // cout << "As many nodes are out before graph construction " << counter << 
   //     ". And total number of nodes is " << n_of_nodes << "." << endl;
    ifstream file_reads; 
    number_of_clusters = 0;
    processed_reads = 0;
    file_reads.open(f_reads.c_str());
    set <int> initialized_cluster;
    cout << "Phase I. Clustering." <<
        " This step is time-consuming." << endl;
    while (findUnassignedRead(initialized_cluster,file_reads))
        growingClsuter(initialized_cluster); 
    file_reads.close();
    outputClusters();
}











